import yaml
from absql.functions import default_constructors


def scalar_to_value(scalar, constructor_dict):
    """
    Converts a YAML ScalarNode to its underlying Python value
    """
    type = scalar.tag.split(":")[-1]
    val = scalar.value
    if isinstance(scalar, yaml.MappingNode):
        val = node_converter(scalar, constructor_dict)
        func = constructor_dict.get(type)
        return func(**val)
    if isinstance(scalar, yaml.SequenceNode):
        val = node_converter(scalar, constructor_dict)
        func = constructor_dict.get(type)
        return func(*val)
    if type.startswith("!"):
        func = constructor_dict.get(type)
        return func(val)
    # Handle null type - https://yaml.org/type/null.html
    if type == "null":
        return None
    return eval('{type}("""{val}""")'.format(type=type, val=val))


def node_converter(x, constructor_dict):
    """
    Converts YAML nodes of varying types into Python values,
    lists, and dictionaries
    """
    if isinstance(x, yaml.ScalarNode):
        # "I am an atomic value"
        return yaml.load(x.value, yaml.SafeLoader)
    if isinstance(x, yaml.SequenceNode):
        # "I am a list"
        return [scalar_to_value(v, constructor_dict) for v in x.value]
    if isinstance(x, yaml.MappingNode):
        # "I am a dict"
        return {
            scalar_to_value(v[0], constructor_dict): scalar_to_value(
                v[1], constructor_dict
            )
            for v in x.value
        }


def wrap_yaml(func, constructor_dict):
    """Turn a function into one that can be run on a YAML input"""

    def ret(loader, x):
        value = node_converter(x, constructor_dict)

        if value is not None:

            if isinstance(value, list):
                return func(*value)

            if isinstance(value, dict):
                return func(**value)

            return func(value)

        else:
            return func()

    return ret


def generate_loader(extra_constructors=None):
    """Generates a SafeLoader with both default and custom constructors"""

    class Loader(yaml.SafeLoader):
        # ensures a new Loader is returned
        # every time the function is called
        pass

    extra_constructors = extra_constructors or []
    unchecked_constructors = default_constructors.copy()

    if isinstance(extra_constructors, list) and len(extra_constructors) > 0:
        extra_constructors = {
            ("!" + func.__name__): func for func in extra_constructors
        }

    if len(extra_constructors) > 0:
        unchecked_constructors.update(extra_constructors)

    # Ensure all tags start with "!"
    checked_constructors = {}
    for tag, func in unchecked_constructors.items():
        if not tag.startswith("!"):
            tag = "!" + tag
        checked_constructors[tag] = func

    for tag, func in checked_constructors.items():
        Loader.add_constructor(tag, wrap_yaml(func, checked_constructors))
    return Loader



from absql.utils import (
    nested_apply,
    get_function_arg_names,
    partialize_function,
    load_body,
)


def test_simple_apply():
    d = {"a": "a", "b": "b", "c": 5}
    got = nested_apply(d, lambda x: x + "zzz")
    want = {"a": "azzz", "b": "bzzz", "c": 5}
    assert got == want


def test_nested_list():
    d = {"a": "a", "b": "b", "c": [5, "d"]}
    got = nested_apply(d, lambda x: x + "zzz")
    want = {"a": "azzz", "b": "bzzz", "c": [5, "dzzz"]}
    assert got == want


def test_nested_dict():
    d = {"a": "a", "b": "b", "c": {5: "d"}}
    got = nested_apply(d, lambda x: x + "zzz")
    want = {"a": "azzz", "b": "bzzz", "c": {5: "dzzz"}}
    assert got == want


def test_get_function_arg_names():
    def func(a, b):
        return a + b

    got = get_function_arg_names(func)
    want = ["a", "b"]
    assert got == want


def test_partialize_function():
    def simple_func(a, engine):
        return a + engine

    simple_func = partialize_function(simple_func, engine=7)
    got = simple_func(3)
    want = 10
    assert got == want


def test_load_body_frontmatter():
    assert (
        load_body("tests/files/simple.sql")
        == "\nSELECT * FROM {{my_table_placeholder}}\n"
    )


def test_load_body_no_frontmatter():
    assert (
        load_body("tests/files/no_frontmatter.sql")
        == "{% set cols='*' %}\n\nSELECT {{cols}} FROM {{greeting}}\n"
    )



from absql import Runner


def test_loader_always_new():
    runner_a = Runner(extra_constructors={"f": lambda: "a"})
    runner_a_want = "SELECT * FROM a"
    render_a_got_1 = runner_a.render("tests/files/loader_test.sql")
    assert render_a_got_1 == runner_a_want

    runner_b = Runner(extra_constructors={"f": lambda: "b"})
    runner_b_want = "SELECT * FROM b"
    runner_b_got = runner_b.render("tests/files/loader_test.sql")
    assert runner_b_got == runner_b_want

    render_a_got_2 = runner_a.render("tests/files/loader_test.sql")
    assert render_a_got_2 == runner_a_want



import pytest
import datetime
from absql import Runner
from copy import copy


@pytest.fixture(scope="function")
def r_original():
    # dt because I am scared to copy modules
    return Runner(f="foo", dt=datetime)


@pytest.fixture(scope="function")
def r_copy(r_original):
    return copy(r_original)


def test_copy_set_context(r_original, r_copy):
    r_copy.set_context(f="bar")
    assert r_original.render("{{f}}") == "foo"
    assert r_copy.render("{{f}}") == "bar"


def test_copy_set_attr(r_original, r_copy):
    r_original.partial_kwargs = ["a"]
    r_copy.partial_kwargs = ["b"]
    assert r_original.partial_kwargs == ["a"]
    assert r_copy.partial_kwargs == ["b"]

    r_copy.extra_context = {"f": "buzz"}
    assert r_original.render("{{f}}") == "foo"
    assert r_copy.render("{{f}}") == "buzz"
